﻿#region License Information
/* HeuristicLab
 * Copyright (C) Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

#define EXPLICIT_SHAPE

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using NumSharp;
using Tensorflow;
using static Tensorflow.Binding;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic {
  public class TreeToTensorConverter {

    private static readonly TF_DataType DataType = tf.float32;

    public static bool TryConvert(ISymbolicExpressionTree tree, int numRows, Dictionary<string, int> variableLengths,
      bool makeVariableWeightsVariable, bool addLinearScalingTerms,
      out Tensor graph, out Dictionary<Tensor, string> parameters, out List<Tensor> variables) {

      try {
        var converter = new TreeToTensorConverter(numRows, variableLengths, makeVariableWeightsVariable, addLinearScalingTerms);
        graph = converter.ConvertNode(tree.Root.GetSubtree(0));

        parameters = converter.parameters;
        variables = converter.variables;
        return true;
      } catch (NotSupportedException) {
        graph = null;
        parameters = null;
        variables = null;
        return false;
      }
    }

    private readonly int numRows;
    private readonly Dictionary<string, int> variableLengths;
    private readonly bool makeVariableWeightsVariable;
    private readonly bool addLinearScalingTerms;

    private readonly Dictionary<Tensor, string> parameters = new Dictionary<Tensor, string>();
    private readonly List<Tensor> variables = new List<Tensor>();

    private TreeToTensorConverter(int numRows, Dictionary<string, int> variableLengths, bool makeVariableWeightsVariable, bool addLinearScalingTerms) {
      this.numRows = numRows;
      this.variableLengths = variableLengths;
      this.makeVariableWeightsVariable = makeVariableWeightsVariable;
      this.addLinearScalingTerms = addLinearScalingTerms;
    }


    private Tensor ConvertNode(ISymbolicExpressionTreeNode node) {
      if (node.Symbol is Constant) {
        var value = (float)((ConstantTreeNode)node).Value;
        var value_arr = np.array(value).reshape(1, 1);
        var var = tf.Variable(value_arr, name: $"c_{variables.Count}", dtype: DataType);
        variables.Add(var);
        return var;
      }

      if (node.Symbol is Variable/* || node.Symbol is BinaryFactorVariable*/) {
        var varNode = node as VariableTreeNodeBase;
        //var factorVarNode = node as BinaryFactorVariableTreeNode;
        // factor variable values are only 0 or 1 and set in x accordingly
        //var varValue = factorVarNode != null ? factorVarNode.VariableValue : string.Empty;
        //var par = FindOrCreateParameter(parameters, varNode.VariableName, varValue);
        var par = tf.placeholder(DataType, new TensorShape(numRows, variableLengths[varNode.VariableName]), name: varNode.VariableName);
        parameters.Add(par, varNode.VariableName);

        if (makeVariableWeightsVariable) {
          var w_arr = np.array((float)varNode.Weight).reshape(1, 1);
          var w = tf.Variable(w_arr, name: $"w_{varNode.VariableName}", dtype: DataType);
          variables.Add(w);
          return w * par;
        } else {
          return varNode.Weight * par;
        }
      }

      //if (node.Symbol is FactorVariable) {
      //  var factorVarNode = node as FactorVariableTreeNode;
      //  var products = new List<Tensor>();
      //  foreach (var variableValue in factorVarNode.Symbol.GetVariableValues(factorVarNode.VariableName)) {
      //    //var par = FindOrCreateParameter(parameters, factorVarNode.VariableName, variableValue);
      //    var par = tf.placeholder(DataType, new TensorShape(numRows, 1), name: factorVarNode.VariableName);
      //    parameters.Add(par, factorVarNode.VariableName);

      //    var value = factorVarNode.GetValue(variableValue);
      //    //initialConstants.Add(value);
      //    var wVar = (RefVariable)tf.VariableV1(value, name: $"f_{factorVarNode.VariableName}_{variables.Count}", dtype: DataType, shape: new[] { 1, 1 });
      //    //var wVar = tf.Variable(value, name: $"f_{factorVarNode.VariableName}_{variables.Count}"/*, shape: new[] { 1, 1 }*/);
      //    variables.Add(wVar);

      //    products.add(wVar * par);
      //  }

      //  return products.Aggregate((a, b) => a + b);
      //}

      if (node.Symbol is Addition) {
        var terms = node.Subtrees.Select(ConvertNode).ToList();
        return terms.Aggregate((a, b) => a + b);
      }

      if (node.Symbol is Subtraction) {
        var terms = node.Subtrees.Select(ConvertNode).ToList();
        if (terms.Count == 1) return -terms[0];
        return terms.Aggregate((a, b) => a - b);
      }

      if (node.Symbol is Multiplication) {
        var terms = node.Subtrees.Select(ConvertNode).ToList();
        return terms.Aggregate((a, b) => a * b);
      }

      if (node.Symbol is Division) {
        var terms = node.Subtrees.Select(ConvertNode).ToList();
        if (terms.Count == 1) return 1.0f / terms[0];
        return terms.Aggregate((a, b) => a / b);
      }

      if (node.Symbol is Absolute) {
        var x1 = ConvertNode(node.GetSubtree(0));
        return tf.abs(x1);
      }

      if (node.Symbol is AnalyticQuotient) {
        var x1 = ConvertNode(node.GetSubtree(0));
        var x2 = ConvertNode(node.GetSubtree(1));
        return x1 / tf.pow(1.0f + x2 * x2, 0.5f);
      }

      if (node.Symbol is Logarithm) {
        return math_ops.log(
          ConvertNode(node.GetSubtree(0)));
      }

      if (node.Symbol is Exponential) {
        return math_ops.pow(
          (float)Math.E,
          ConvertNode(node.GetSubtree(0)));
      }

      if (node.Symbol is Square) {
        return tf.square(
          ConvertNode(node.GetSubtree(0)));
      }

      if (node.Symbol is SquareRoot) {
        return math_ops.sqrt(
          ConvertNode(node.GetSubtree(0)));
      }

      if (node.Symbol is Cube) {
        return math_ops.pow(
          ConvertNode(node.GetSubtree(0)), 3.0f);
      }

      if (node.Symbol is CubeRoot) {
        return math_ops.pow(
          ConvertNode(node.GetSubtree(0)), 1.0f / 3.0f);
        // TODO
        // f: x < 0 ? -Math.Pow(-x, 1.0 / 3) : Math.Pow(x, 1.0 / 3),
        // g:  { var cbrt_x = x < 0 ? -Math.Pow(-x, 1.0 / 3) : Math.Pow(x, 1.0 / 3); return 1.0 / (3 * cbrt_x * cbrt_x); }
      }

      if (node.Symbol is Sine) {
        return tf.sin(
          ConvertNode(node.GetSubtree(0)));
      }

      if (node.Symbol is Cosine) {
        return tf.cos(
          ConvertNode(node.GetSubtree(0)));
      }

      if (node.Symbol is Tangent) {
        return tf.tan(
          ConvertNode(node.GetSubtree(0)));
      }

      if (node.Symbol is Mean) {
        return tf.reduce_mean(
          ConvertNode(node.GetSubtree(0)),
          axis: new[] { 1 },
          keepdims: true);
      }

      //if (node.Symbol is StandardDeviation) {
      //  return tf.reduce_std(
      //    ConvertNode(node.GetSubtree(0)),
      //    axis: new [] { 1 }
      // );
      //}

      if (node.Symbol is Sum) {
        return tf.reduce_sum(
          ConvertNode(node.GetSubtree(0)),
          axis: new[] { 1 },
          keepdims: true);
      }

      if (node.Symbol is StartSymbol) {
        Tensor prediction;
        if (addLinearScalingTerms) {
          // scaling variables α, β are given at the beginning of the parameter vector
          var alpha_arr = np.array(1.0f).reshape(1, 1);
          var alpha = tf.Variable(alpha_arr, name: "alpha", dtype: DataType);
          var beta_arr = np.array(0.0f).reshape(1, 1);
          var beta = tf.Variable(beta_arr, name: "beta", dtype: DataType);
          variables.Add(alpha);
          variables.Add(beta);
          var t = ConvertNode(node.GetSubtree(0));
          prediction = t * alpha + beta;
        } else {
          prediction = ConvertNode(node.GetSubtree(0));
        }

        return tf.reduce_sum(prediction, axis: new[] { 1 });
      }

      throw new NotSupportedException($"Node symbol {node.Symbol} is not supported.");
    }

    public static bool IsCompatible(ISymbolicExpressionTree tree) {
      var containsUnknownSymbol = (
        from n in tree.Root.GetSubtree(0).IterateNodesPrefix()
        where
          !(n.Symbol is Variable) &&
          //!(n.Symbol is BinaryFactorVariable) &&
          //!(n.Symbol is FactorVariable) &&
          !(n.Symbol is Constant) &&
          !(n.Symbol is Addition) &&
          !(n.Symbol is Subtraction) &&
          !(n.Symbol is Multiplication) &&
          !(n.Symbol is Division) &&
          !(n.Symbol is Logarithm) &&
          !(n.Symbol is Exponential) &&
          !(n.Symbol is SquareRoot) &&
          !(n.Symbol is Square) &&
          !(n.Symbol is Sine) &&
          !(n.Symbol is Cosine) &&
          !(n.Symbol is Tangent) &&
          !(n.Symbol is HyperbolicTangent) &&
          !(n.Symbol is Erf) &&
          !(n.Symbol is Norm) &&
          !(n.Symbol is StartSymbol) &&
          !(n.Symbol is Absolute) &&
          !(n.Symbol is AnalyticQuotient) &&
          !(n.Symbol is Cube) &&
          !(n.Symbol is CubeRoot) &&
          !(n.Symbol is Mean) &&
          //!(n.Symbol is StandardDeviation) &&
          !(n.Symbol is Sum)
        select n).Any();
      return !containsUnknownSymbol;
    }
  }
}
